# -*- coding: utf-8 -*-


class cycle(object):
    """
    Simple circular buffer generator that doesn't require the storing of a full
    iterable like `collections.cycle`.

    WARNING: This object is an endless generator.  Be careful to not
    boundlessly iterate.
    """
    def __init__(self, start=0, stop=None, step=1):
        if stop is None:
            stop = start
            start = 0

        self.start = start
        self.step = step
        self.stop = stop - self.step
        # to mirror range behavior, we want to stop one step so we never
        # increment to >= stop.

        # init _current to 1 step behind so the first value is the actual
        # start after increment
        self._current = self.start - self.step

    def __iter__(self):
        return self

    def __next__(self):
        # if we aren't at the bound...
        if self._current < self.stop:
            # ...step
            self._current += self.step
        else:
            # we are at the bound...so start over
            self._current = self.start

        return self._current



import multiprocessing as mp
import queue as q
import sys
import uuid

from coppyr.singleton import Singleton


class Worker(object):
    """
    This is a utility wrapper for taking a function and converting it into a
    callable that accepts Queue pipes from Parent process rather than direct
    inputs.
    """
    def __init__(self, func, input_type="args", name=None, logger=None):
        self.func = func
        self.input_type = input_type
        self.name = name or str(uuid.uuid4())[:8]
        self.logger = logger

    def __call__(self, input_queue, result_queue):
        # disable logging from dependencies as logging in Python is not process safe
        import logging
        logging.disable(logging.CRITICAL)

        while True:
            # get an input
            try:
                inp = input_queue.get(timeout=10)
            except q.Empty as e:
                if self.logger is not None:
                    self.logger.info(f"[parallel.Worker {self.name}] Queue get timed out")
                continue

            # This will block until an item is retrieved from the queue
            # the queue.
            # https://docs.python.org/3.6/library/multiprocessing.html#multiprocessing.Queue

            # to kill a process, put None into the queue
            if inp is None:
                # print(f'Closing worker {self.name}')
                break
            try:
                if self.input_type == "arg":
                    result = self.func(inp)
                elif self.input_type == "args":
                    result = self.func(*inp)
                elif self.input_type == "kwargs":
                    result = self.func(**inp)
                elif self.input_type == "mixed":
                    result = self.func(
                        *inp.get("args", []),
                        **inp.get("kwargs", {})
                    )
                else:
                    raise TypeError(f"Unrecognized input type \"{self.input_type}\"")
            except Exception as e:
                _, _, tb = sys.exc_info()
                result = e.with_traceback(tb)
                result_queue.put(result)
            else:
                result_queue.put(result)


class Pool(object):
    """
    Class that spawns a group of workers and can send work / receive results
    from that group.  This allows us to keep workers alive (and their state)
    in between workloads.

    BUYER BE WARNED...This module worked at one point quite well for farming
    out local processing, but recently when trying to make use of it in a
    Google Function inter process communication was very unstable.  There is
    likely some work to be done to shore this utility area up before it is
    ready for prime time.

    Also, consider a rework to use concurrent.Futures as an alternative now in
    Python 3.  Under the covers that stdlib uses multiprocessing primatives for
    Process executors, but perhaps it is a bit more user friendly.
    """
    def __init__(self, func, size=4, input_type="args", manager=None,
                 name=None, logger=None):
        # https://stackoverflow.com/questions/43439194/python-multiprocessing-queue-vs-multiprocessing-manager-queue/45236748#45236748
        # https://docs.python.org/3/library/multiprocessing.html#multiprocessing-managers
        if manager is None:
            # hold reference open to prevent gc if self-managed
            self.manager = mp.Manager()
            # If we don't do this then the manager might be GC'd at any time
            # and once that happens its queue proxies will not be usable.  We
            # don't have to do it for externally passed managers because we
            # implicitly trust that instantiating logic will hold open a
            # reference to prevent GC for us.

            manager = self.manager

        self.sent = manager.Value(int, 0)
        self.input_queue = manager.Queue()
        self.result_queue = manager.Queue()
        self.name = name or str(uuid.uuid4())[:8]
        self.logger = logger

        self.workers = [
            mp.Process(
                target=Worker(func, input_type, name=f"{self.name}_{i}", logger=self.logger),
                args=(self.input_queue, self.result_queue),
                daemon=True,
                name=f"{self.name}_{i}"
            )
            for i in range(size)
        ]

        # start the workers
        for process in self.workers:
            process.start()

    @property
    def waiting(self):
        """
        Boolean indicator showing whether or not there is still work queued.
        """
        return not self.input_queue.empty()

    @property
    def empty(self):
        """
        Boolean indicator showing whether or not there are result messages
        available.
        """
        return self.result_queue.empty()

    @property
    def idle(self):
        """
        Use our self tracked message counter to indicate whether or not the
        pool has outstanding work, doing work, or has results that haven't been
        handled.
        """
        return self.sent.value == 0

    def send(self, *args):
        """
        Send tasks to the pool.  Returns the count of tasks sent.
        """
        current = 0
        for inp in args:
            self.input_queue.put(inp)
            self.sent.value += 1
            current += 1

        return current

    def recv(self, timeout=None):
        result = self.result_queue.get(timeout=timeout)
        self.sent.value -= 1
        return result

    def results(self, timeout=None):
        """
        Read results from children, this will block until we receive responses
        for each message sent
        """
        results = []

        for _ in range(self.sent.value):
            try:
                results.append(self.recv(timeout=timeout))
            except q.Empty as e:
                results.append(e)
                break

        return results

    def run(self, *args):
        # if there is outstanding work in the input queue, don't execute since
        # we can't guarantee the results are from this run call
        if self.sent.value != 0:
            return []

        self.send(*args)

        # return results
        return self.results()

    def stop(self):
        # clean up workers (in case signaled close didn't get all of them)
        # print('Terminating worker processes')
        for process in self.workers:
            if process.is_alive():
                process.terminate()

        # print('Resetting worker pool and queues')
        self.workers = []  # just for good measure deallocate for GC

        # This isn't strictly necessary, but making sure that workers are
        # cleaned up before the queues will avoid raising EOF errors in
        # workers.  (EOF happens if queue is closed before the workers exit)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # https://docs.python.org/3/reference/datamodel.html#object.__exit__
        self.stop()

    def __del__(self):
        # NOTE: __del__ is not reliably called by the GC so anything that we
        # want to make SURE runs shouldn't go here.
        self.stop()


class LogHandler(object):
    """
    Alternative worker implementation that initializes logging in the forked
    process and handles remote log structure message from the Logging pool.
    """
    def __call__(self, input_queue, result_queue):
        from coppyr.context import Context
        context = Context()
        logger = context.log

        while True:
            try:
                # get an input
                inp = input_queue.get(timeout=5)
            except q.Empty as e:
                continue

            # handle special increment message
            if isinstance(inp, str) and inp == "inc_action_id":
                context.inc_action_id()
                result_queue.put(None)
                continue

            level, msg = inp.get("args", ("error", "Malformed remote log payload"))
            kwargs = inp.get("kwargs", {})

            try:
                # log the message (e.g. logger.error("my message", exc_info=True))
                getattr(logger, level)(msg, **kwargs)
            except AttributeError:
                logger.error(f"Invalid logging level: \"{level}\"")

            # just send an empty result to indicate message handling
            result_queue.put(None)


class Logger(Pool, Singleton):
    """
    A multiprocessing-safe logger.  This is a process pool which spawns a
    (SINGLE) separate process that will be responsible for log handling.  Log events are
    converted into tuples which are then logged by the separate process.
    """
    def __init__(self, manager, level):
        # manager is required because we need this pool to be picklable
        self.level = level

        # basically init an empty pool
        super().__init__(lambda: True, size=0, manager=manager)

        self.workers.append(
            mp.Process(
                target=LogHandler(),
                args=(self.input_queue, self.result_queue),
                daemon=True,
                name="parallel.logger"
            )
        )

        # start the logger
        self.workers[-1].start()

    def inc_action_id(self):
        """
        Send special message which increments the remote action id
        """
        self.send("inc_action_id")

    def critical(self, msg, **kwargs):
        self.send({"args": ("critical", msg), "kwargs": kwargs})

    def error(self, msg, **kwargs):
        self.send({"args": ("error", msg), "kwargs": kwargs})

    def warning(self, msg, **kwargs):
        self.send({"args": ("warning", msg), "kwargs": kwargs})

    def info(self, msg, **kwargs):
        self.send({"args": ("info", msg), "kwargs": kwargs})

    def debug(self, msg, **kwargs):
        self.send({"args": ("debug", msg), "kwargs": kwargs})


def run(func, inputs, n=4, input_type="args"):
    """
    Take a function and parallelize it.  This is a thin utility for
    experimentation purposes.

    :param func: Callable
        Function to call in child processes.
    :param inputs:  Iterable[Tuple|List|Dict]
        Iterable of inputs that you want to pass to the remote functions.
    :param n: Integer
        Number of children you want to spawn.
    :param input_type: String ["arg" | "args" | "kwargs"]
        Whether or not the passed inputs are args (ordered) or kwargs (key-word
        arguments).  This dictates how the inputs are expanded when calling the
        function.
    :return: List
        List of results that were returned by child workers.
    """

    exec_pool = Pool(func, size=n, input_type=input_type)

    # send the work to the Pool
    results = exec_pool.run(*inputs)

    exec_pool.stop()

    # return results
    return results



# -*- coding: utf-8 -*-
import subprocess
from logging import Logger

from typing import ByteString, List, Optional, Tuple, TextIO, Union

from coppyr.error import CoppyrError


class CoppyrSubpError(CoppyrError):
    description = "An error occurred durring subp.call"


def call(
    command: str,
    check: bool=True,
    raw: bool=False,
    fg: bool=False,
    input: Optional[str]=None,
    stdin: Optional[TextIO]=None,
    stdout: Optional[TextIO]=None,
    stderr: Optional[TextIO]=None,
    log: Optional[Logger]=None
) -> Union[
    Tuple[int, List[str], List[str]],
    Tuple[int, ByteString, ByteString]
]:
    """
    Calls subprocess.run with the pass command.

    :param command: String
        Full shell command
    :param check: Boolean
        Check the return code or not
    :param raw: Boolean
        Flag to return raw outputs or not
    :param fg: Boolean
        Flag to indicate whether or not the command should be run in
        the foreground.
    :param input: String
        String input to send via stdin/communicate to the process.
    :param stdin: File-like object
        stdin to passthrough to the subprocess.  Default is
        subprocess.PIPE.
    :param stdout: File-like object
        stdout to passthrough to the subprocess.  Default is
        subprocess.PIPE.
    :param stderr: File-like object
        stderr to passthrough to the subprocess.  Default is
        subprocess.PIPE.
    :param log: logging.Logger
        Logger to send log entries to.  Can be `None` to disable logging.
        Disabled by default.
    :return: retcode int, stdout [], stderr []
    """
    subprocess_params = dict(
        shell=True
    )

    if not fg:
        stdin = subprocess.PIPE if stdin is None else stdin  # type: ignore
        stdout = subprocess.PIPE if stdout is None else stdout  # type: ignore
        stderr = subprocess.PIPE if stderr is None else stderr  # type: ignore

        subprocess_params.update(dict(  # type: ignore
            stdin=stdin,
            stdout=stdout,
            stderr=stderr
        ))

    if input:
        if 'stdin' in subprocess_params:
            del subprocess_params['stdin']
        subprocess_params['input'] = input  # type: ignore

    if log is not None:
        log.debug(f"coppyr.subp.call: Calling \"{command}\"")

    process = subprocess.run(command, **subprocess_params)

    try:
        if process.returncode != 0 and check:
            if log is not None:
                log.error(
                    f"piston.subp.call: \"{command}\" returned a non-zero "
                    f"exit code \"{process.returncode}\""
                )

            raise CoppyrSubpError(
                message=f"\"{command}\" returned a non-zero exit code.",
                payload=dict(returncode=process.returncode)
            )
        elif fg:
            return process.returncode, [], []
        elif raw:
            return process.returncode, process.stdout, process.stderr
        else:
            # out, err are bytes objects
            out = process.stdout.decode("utf-8").split('\n')
            err = process.stderr.decode("utf-8").split('\n')

            if log is not None:
                if out:
                    log.debug("coppyr.subp.call: stdout:\n" + "\n".join(out) + "\n")
                if process.stderr:
                    log.error("coppyr.subp.call: stderr:\n" + "\n".join(err) + "\n")

            return process.returncode, out, err
    except CoppyrSubpError:
        raise
    except Exception as e:
        if log is not None:
            log.error(
                f"coppyr.subp.call: Caught \"{e.__class__.__name__}\" during "
                "subp.call."
            )

        raise CoppyrSubpError(
            f"Caught \"{e.__class__.__name__}\" during subp.call.",
            payload={"command": command},
            caught=e
        )



# -*- coding: utf-8 -*-
import inspect

from typing import Any, Iterable, List, Optional, Type, Union


def get_slots(
    slotted_obj: Union[Type[Any], Any],
    ignored: Optional[Iterable]=None
) -> List[str]:
    """
    Simple recursive logic that will traverse a (slotted) object's inheritance
    graph to retrieve a complete set of slotted attributes.
    """
    slots = set(slotted_obj.__slots__)  # type: ignore

    slotted_cls = slotted_obj \
        if inspect.isclass(slotted_obj) \
        else slotted_obj.__class__

    for base_cls in slotted_cls.__bases__:
        if hasattr(base_cls, "__slots__"):
            slots.update(get_slots(base_cls))

    return list(slots) if ignored is None \
        else [i for i in slots if i not in ignored]




# -*- coding: utf-8 -*-
import pkgutil
import importlib
import inspect

from types import ModuleType
from typing import Any, List, Type


def import_object(obj_path: str) -> Type[Any]:
    """
    Helper function for importing a single object from another module.

    :obj_path: String
        Full path to the object (e.g. "coppyr.errors.CoppyrError").
    :return: Object
        The imported object.
    """
    path_parts = obj_path.split(".")
    module_name, obj_name = ".".join(path_parts[:-1]), path_parts[0]
    module = import_module(module_name)
    return getattr(module, obj_name)


def import_module(mod_path: str) -> ModuleType:
    """
    Just a local wrapper for `importlib.import_module`

    :mod_path: String
        Full path to the module (e.g. "piston.errors").
    :return: Module
    """
    return importlib.import_module(mod_path)


def import_subclasses(
    module, baseclass: Type[Any],
    recursive: bool=False,
    include_base: bool=False
) -> List[ModuleType]:
    """
    Function to import all classes from a module that are subclasses of a
    desired class.

    This is useful when designing an extension system where extension objects
    inherit from a common base.  This function allows dynamic loading by the
    runtime rather than imperative registration of extensions.

    :module: String/Module
        Target module either qualfied string name or an already imported module
        object.
    :baseclass: Class
        Class definition to use as a a base class.
    :recursive: Boolean
        Whether or not to check and load from all sub-modules.
    :include_base: Boolean
        Whether or not to include instances of the baseclass itself in the
        result.
    """
    results = []

    module = import_module(module) if isinstance(module, str) else module
    prefix = module.__name__ + "."  # "piston.errors."

    for _, name, ispkg in pkgutil.iter_modules(module.__path__):
        location = prefix + name

        if ispkg:
            # if it is another module...
            if recursive:
                # ...and recursive is enabled, then recurse
                results += import_subclasses(
                    name,
                    baseclass=baseclass,
                    recursive=recursive,
                    include_base=include_base
                )
        else:
            # load the module (python file)
            module = import_module(location)

            # Walk the objects in the module
            for obj in module.__dict__.values():
                # If it is a class definition
                if inspect.isclass(obj):
                    # If it is a subclass of the baseclass
                    if issubclass(obj, baseclass):
                        # If we are supposed to ignore base and this obj has
                        # the same __name__, ignore it.
                        if not include_base \
                                and obj.__name__ == baseclass.__name__:
                            continue

                        # otherwise add the found class defintion
                        results.append(obj)

    return results



# -*- coding: utf-8 -*-
import abc
from typing import Dict

from coppyr.collections import DotDict
from coppyr.error import CoppyrError


# errors


class CoppyrConfigIOError(CoppyrError):
    description = "Error opening the config file."


class CoppyrConfigYAMLError(CoppyrError):
    description = "Couldn't load YAML."


# objects


class AbstractConfig(DotDict):
    def __init__(self, file_path: str=None):
        """
        :param file_path: String
            Path to the YAML file to load.  This can be absolute or relative.
        """
        if file_path is None:
            raise CoppyrConfigIOError(
                "File path not specified."
            )

        self.__file_path__ = file_path

        # Load KV arguments into DotDict
        super().__init__(**self.load(file_path))

    @abc.abstractstaticmethod
    def load(file_path: str) -> Dict:
        return {}


class YAMLConfig(AbstractConfig):
    @staticmethod
    def load(file_path: str):
        import yaml

        try:
            with open(file_path, "r") as f:
                raw = yaml.load(f, Loader=yaml.CLoader)
        except IOError as e:
            raise CoppyrConfigIOError(
                f"Failed to load file path \"{file_path}\".",
                caught=e
            )
        except Exception as e:
            raise CoppyrConfigYAMLError(caught=e)

        return raw


class TOMLConfig(AbstractConfig):
    @staticmethod
    def load(file_path: str):
        import toml

        try:
            raw = toml.load(file_path)
        except IOError as e:
            raise CoppyrConfigIOError(
                f"Failed to load file path \"{file_path}\".",
                caught=e
            )
        except Exception as e:
            raise CoppyrConfigYAMLError(caught=e)

        return raw


# TODO: Add an environment lookup for the configuration bit.



import time

from daemon.runner import DaemonRunner

from coppyr.context import Context


# When running in a daemon context we _need_ context to be setup to
# appropriately pass the right values to os.fork().
context = Context()
context.setup()


class Runner(DaemonRunner):
    # See source code here: https://pagure.io/python-daemon/
    def __init__(self, app):
        super().__init__(app)

        self.app = app

        # daemon context modifications
        # https://pagure.io/python-daemon/blob/main/f/daemon/daemon.py#_108
        self.daemon_context.detach_process = getattr(
            self.app, "detach_process", None
        )
        self.daemon_context.files_preserve = getattr(
            self.app,
            "files_preserve",
            context.get_file_handlers()
        )
        self.daemon_context.signal_map = getattr(self.app, "signal_map", {})

    def _open_streams_from_app_stream_paths(self, app):
        # Ugly hack to fix a bug where DaemonRunner can't open stdin and stdout
        # in "w+t" mode (text w+).  This seems to be a Python 3 thing...see
        # more information here: https://bugs.python.org/issue20074
        self.daemon_context.stdin = open(app.stdin_path, "rt")
        self.daemon_context.stdout = open(app.stdout_path, "w+b", buffering=0)
        self.daemon_context.stderr = open(app.stderr_path, "w+b", buffering=0)


class App:
    def __init__(self):
        super().__init__()

        global context
        self.context = context

        # daemon app attributes
        self.pidfile_path = f"/tmp/{self.context.app_name}.pid"
        self.pidfile_timeout = 5
        self.stdin_path = "/dev/null"
        self.stdout_path = "/dev/stdout"
        self.stderr_path = "/dev/stderr"

        # daemon context settings
        self.detach_process = None
        self.files_preserve = context.get_file_handlers()
        self.signal_map = {}
        # Example signal handling:
        # self.daemon_context.signal_map = {
        #     signal.SIGTERM: callable<(signum, frame)>
        # }
        # See IRL example here:
        # https://github.com/nginxinc/nginx-amplify-agent/blob/master/amplify/agent/common/runner.py#L25

    def run(self):
        while True:
            print("Default daemon is running...sleeping")
            time.sleep(60)



# -*- coding: utf-8 -*-
import json
from typing import Type


class CoppyrError(Exception):
    """
    Simple base exception to provide boiler plate for readable, consistent
    error messages.  Also has a dict representation for easy(ish) conversion to
    JSON for web use cases.

    Provides some basic extensions and representation handling.
    """

    description = "Something unexpected happened."

    def __init__(self, message=None, payload=None, caught=None):
        super().__init__()
        self.message = message if message is not None else self.description
        self.payload = payload if payload is not None else {}

        if caught is not None:
            self.payload.update({
                "caught": {
                    "name": caught.__class__.__name__,
                    "message": getattr(caught, "message", str(caught)),
                }
            })

    def to_dict(self):
        return {
            "error": self.__class__.__name__,
            "message": self.message,
            "payload": self.payload if self.payload is not None else {}
        }

    def to_json(self, indent=None):
        return json.dumps(self.to_dict(), indent=indent)

    def __repr__(self):
        return f"{self.__class__.__name__}"\
               f"(message={self.message}, payload={self.payload})"

    def __str__(self):
        return self.message


def exc_formatter(
    e: Exception,
    coerce: bool = False,
    cls: Type[CoppyrError] = CoppyrError
):
    if coerce:
        e = cls(caught=e)

    import sys
    _, _, tb = sys.exc_info()

    error_with_tb = e.with_traceback(tb)
    return error_with_tb



# -*- coding: utf-8 -*-
import logging

from coppyr import Context


# Module variables


LOGGERS_CACHE = {}

# Some example, basic formats.
COPPYR_FMT = "%(asctime)s [%(action_id)s] %(levelname)s \"%(message)s\""
COPPYR_FMT_WITH_TIME = "%(asctime)s [%(action_id)s] %(levelname)s " \
                       "\"%(message)s\" %(action_duration_str)s"

# Example/Default logging config
DEFAULT_CONFIG = {
    "version": 1,
    "formatters": {
        "coppyr": {
            "format": COPPYR_FMT
        },
        "coppyr-with-time": {
            "format": COPPYR_FMT_WITH_TIME
        }
    },
    "handlers": {
        "file-default": {
            "class": "logging.FileHandler",
            "level": "DEBUG",
            "formatter": "coppyr-with-time",
            "filename": f"/tmp/{Context().app_name}.log",
            "mode": "a"
        },
        "file-devnull": {
            "class": "logging.FileHandler",
            "level": "DEBUG",
            "formatter": "coppyr-with-time",
            "filename": "/dev/null",
            "mode": "a"
        },
        "stream-stdout": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "coppyr-with-time",
            "stream": "ext://sys.stdout"
        }
    },
    "loggers": {
        "file": {
            "level": "DEBUG",
            "qualname": "default",
            "handlers": ["file-default"],
            "propagate": False
        },
        "devnull": {
            "level": "DEBUG",
            "qualname": "devnull",
            "handlers": ["file-devnull"],
            "propagate": False
        },
        "stdout": {
            "level": "DEBUG",
            "qualname": "stdout",
            "handlers": ["stream-stdout"],
            "propagate": False
        }
    },
    "root": {
        "level": "DEBUG",
        "formatter": "coppyr",
        "handlers": ["stream-stdout"]  # change this to file-default to have all
                                       # logs appear in the log file (including
                                       # logs from dependency libs)
    }
}


class CoppyrLogRecord(logging.LogRecord):
    """
    This log record class can consult `coppyr.Context` for more variables after
    normal `logging.LogRecord` init.  This is particularly useful for rich log
    contexts in complex services.

    The keys that are included/retrieved are configurable by modifying the
    `context.logging_keys` list.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        context = Context()
        for log_key in context.logging_keys:
            setattr(self, log_key, getattr(context, log_key))


class CoppyrLogger(logging.getLoggerClass()):
    """
    This is a custom logger class that simply users the custom log record above
    instead of the default one.
    """
    def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None,
                   *arg, **kwargs):
        return CoppyrLogRecord(name, level, fn, lno, msg, args, exc_info, func)


def setup(dict_config=None, logger_class=CoppyrLogger):
    """
    Simple wrapper that accepts a dictionary and passes it to
    `logging.config.dictConfig`.

    WARNING: This method should only be called once as multiple calls to
    `logging.config.dictConfig` can cause errors.

    :param dict_config: Dict
        Dictionary following the configuration format of the logging built-in.
        For reference consult:
            https://docs.python.org/3/library/logging.config.html
    """
    if dict_config is None:
        dict_config = DEFAULT_CONFIG

    # Set the logging lib to use our logger and log record objects.
    logging.setLoggerClass(logger_class)

    from logging.config import dictConfig
    dictConfig(dict_config)


def get(name="file", level=None):
    """
    Creates logger object to specified log and caches it in `LOGGERS_CACHE`
    dict.

    :param name: String
        Log name.  `None` returns the "root" logger from `logging.getLogger()`.
    :param level: String
        Log level to set on logger. `None` uses the default log setting.
    :return: CoppyrLogger
        Logger object.
    """
    # If logger is not in cache, create a new one...
    if name not in LOGGERS_CACHE:
        logger = logging.getLogger(name)

        # Set log level if applicable.
        if level is not None:
            logger.setLevel(level)
        # TODO: We should sanity check level value first.

        # Cache logger.
        LOGGERS_CACHE[name] = logger

    return LOGGERS_CACHE[name]


def shutdown():
    """
    Call into `logging` to perform an orderly shutdown by flushing and closing
    all handlers.
    """
    logging.shutdown()



# -*- coding: utf-8 -*-
from coppyr.error import CoppyrError


class CoppyrWebError(CoppyrError):
    code = 500

    def to_dict(self):
        result = super().to_dict()
        result.update(code=self.code)
        return result


def multidict_to_dict(multidict):
    """
    Takes a Werkzeug.Multidict and converts it to a regular dict.

    http://werkzeug.pocoo.org/docs/0.12/datastructures/#werkzeug.datastructures.MultiDict

    Note: In newer versions of Werkzeug this converter might be unnecessary.
    """
    return {
        k: multidict.getlist(k)[0] if len(multidict.getlist(k)) == 1
        else multidict.getlist(k)
        for k in multidict.keys()
    }


def strip_parser_none(kwargs):
    """
    Helper to remove all "None" values from a reqparser.parse_args().

    Note: This helper leverages pass by reference in order to communicate
    results to calling context.
    """
    _ = []
    for k, v in kwargs.items():
        if v is None:
            _.append(k)
    for k in _:
        del kwargs[k]



# -*- coding: utf-8 -*-
import os
import io
import itertools
import shutil
import sys

from typing import List, Dict

from setuptools import Command
from pip._internal.req import parse_requirements as pip_parse_requirements


def get_readme(fname="README.md", path=os.getcwd()):
    path = os.path.join(path, fname)
    with io.open(path, encoding="utf-8") as f:
        return "\n" + f.read()


class UploadCommand(Command):
    """
    Support setup.py upload.

    NOTE: This upload command requires `twine` to upload to PyPI.  See an
    example usage via NGINX Crossplane:

    https://github.com/nginxinc/crossplane/blob/master/setup.py#L92
    """

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print(f'\033[1m{s}\033[0m')

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds...")
            shutil.rmtree(os.path.join(os.getcwd(), "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution...")
        os.system(f"{sys.executable} -m build")

        self.status("Uploading the package to PyPI via Twine...")
        os.system("twine upload dist/*")

        sys.exit()


def parse_requirements(fpath: str) -> List[str]:
    requirements = []

    for req in pip_parse_requirements(fpath, session="tmp"):
        if hasattr(req, "req"):
            requirements.append(str(req.req))
        else:
            requirements.append(str(req.requirement))

    return requirements


def parse_extras(**kwargs: str) -> Dict[str, List[str]]:
    """
    Parse a key-value pairing of extras keys and file paths and return a dict
    which can be passed to setup.py as "extras_require".

    This logic will automatically include an "all" extras argument which will
    include all extras parsed by this func.  It will exclude the "dev" key from
    this "all" tag.

    https://setuptools.pypa.io/en/latest/userguide/dependency_management.html#optional-dependencies
    """
    extras_dict = {
        k: parse_requirements(v)
        for k, v in kwargs.items()
    }

    extras_dict["all"] = list(
        itertools.chain(
            *[v for k, v in extras_dict.items() if k != "dev"]
        )
    )

    return extras_dict



# -*- coding: utf-8 -*-
from coppyr.collections import DotDict


class Singleton(object):
    """
    This object will only ever create one instance of itself in (interpreter
    local) memory.  Future constructions will just return the previously
    constructed instance.

    WARNING: Since construction is separate from initialization (__new__
    vs. __init__), it is possible that the __init__ method can inadvertantly be
    called multiple times for a single instance.  While new will return the
    previously instantiated object, the interpreter will flow through __init__
    as if it had just constructed a fresh object.
    """
    _instance = None
    _init = True  # use this flag to skip future init calls if desirable

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        self._init = False


class Namespace(Singleton, DotDict):
    """
    Simple object that acts as a an easily shared variable store.
    """

    def __init__(self, *args, **kwargs):
        if not self._init:
            return

        super().__init__(*args, **kwargs)



# -*- coding: utf-8 -*-


class lazyproperty(object):
    """
    Decorator for creating properties that're only evaluated once.
    """
    def __init__(self, method):
        self.method = method

    def __get__(self, instance, cls):
        value = self.method(instance)
        setattr(instance, self.method.__name__, value)
        return value



# -*- coding: utf-8 -*-
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from coppyr.types import slot


def getattrs(obj: Any) -> List[str]:
    if hasattr(obj, "__slots__"):
        return slot.get_slots(obj)
    else:
        return list(obj.__dict__.keys())


def _diff_tuple(
    left: Any,
    right: Any,
    invert: bool=False
) -> Union[Tuple[Any, Any], None]:
    """
    Basic, atomic diff unit that returns a common tuple format.

    :param left: Any
        Left value
    :param right: Any
        Right value
    :param invert: Boolean
        Whether or not to reverse the left and right treatment of the passed
        values.  This is a convenience parameter so that actual diff logic can
        have consistent handling of tuple construction independent of desired
        orientation.
    :return: Tuple/sub-diff
    """
    if left == right:
        return None
    else:
        return (left, right) if not invert else (right, left)


def _diff_none(
    obj: Any,
    attrs: Optional[Iterable[str]]=None,
    invert: bool=False
) -> Dict[str, Tuple[Any, Any]]:
    """
    If an object is being diffed with `None` then the object is entirely "new"
    so we should return a complete diff of all targeted attributes.
    """
    result = {}

    attrs = attrs if attrs is not None else getattrs(obj)

    for attr in attrs:  # type: ignore
        obj_attr = getattr(obj, attr)
        if isinstance(obj_attr, (list, tuple)):
            result[attr] = [
                _diff_tuple(x, None, invert=invert) for x in obj_attr
            ]
        else:
            result[attr] = _diff_tuple(obj_attr, None, invert=invert)

    return result


def diff(
    left_obj: Any,
    right_obj: Any,
    attrs: Optional[Iterable[str]]=None,
    invert: bool=False
) -> Dict[str, Tuple[Any, Any]]:
    """
    Diff two objects and get a diff structure back which shows left vs. right
    changes.  From this object you can see what things are different, what
    things are added, and what things were removed.

    :param left_obj: Any
    :param right_obj: Any
    :param attrs: Iterable of Strings
        Attribute names of the objects to include in the diff.  Will default to
        __dict__ and/or __slots__ keys.
    :param invert: Boolean
        Whether or not reverse the diff direction.
    :return: Dict[str, Any]
        Attribute map to diff tuples.
    """
    # re-orient
    if invert:
        left_obj, right_obj = right_obj, left_obj
        invert = False

    # If attrs is not explicitly specified, infer them by getting the superset
    # of both object attributes
    if attrs is None:
        attrs = set(getattrs(left_obj))
        attrs = list(attrs.union(getattrs(right_obj)))

    # `None` is a special case, so handle it first
    if left_obj is None and right_obj is None:
        return {}
    if right_obj is None:
        return _diff_none(left_obj, attrs)
    if left_obj is None:
        return _diff_none(right_obj, attrs, invert=True)

    # Check that object types match
    if not issubclass(right_obj.__class__, left_obj.__class__) \
            or left_obj.__class__.__name__ != right_obj.__class__.__name__:
        raise ValueError('Object type mismatch')

    # start diff
    result = {}

    for attr in attrs:
        left = getattr(left_obj, attr)
        right = getattr(right_obj, attr)

        # if they are equal, no diff to show
        if left == right:
            continue
        else:
            result[attr] = _diff_tuple(left, right, invert=invert)

    return result



# -*- coding: utf-8 -*-
import os

from .context import Context
from .error import CoppyrError


__title__ = "coppyr"
__summary__ = "A collection of boilerplate for Python applications."
__url__ = "https://github.com/gshulegaard/coppyr"

__version__ = "0.9.3"

__author__ = "Grant Hulegaard"
__email__ = "loki.labrys@gmail.com"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2019 Grant Hulegaard"


here = os.path.dirname(__file__)



# -*- coding: utf-8 -*-


class DotDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # recursively cast dicts to DotDicts that may have been set by init
        for k, v in self.items():
            if isinstance(v, dict):
                self[k] = DotDict(**v)

    def __getattr__(self, k):
        if k.startswith("__"):
            return super().__getattr__(k)

        return super().get(k)

    def __setattr__(self, k, v):
        if k.startswith("__"):
            return super().__setattr__(k, v)

        super().__setitem__(k, v)

    def __delattr__(self, k):
        super().__delitem__(k)

    def __setitem__(self, k, v):
        # support recursive dot spec
        if isinstance(v, dict) and not isinstance(v, DotDict):
            v = DotDict(**v)

        super().__setitem__(k, v)



# -*- coding: utf-8 -*-
import os
import sys
import time

from typing import Optional

from coppyr.types import lazyproperty, Singleton, Namespace
from coppyr.collections import cycle, DotDict


# TODO: These sys variables should really be applied elsewhere.
sys.tracebacklimit = 10000
sys.setrecursionlimit(2048)


class TimeKeeper:
    def __init__(self):
        self.timings = []
        self.relative_base = time.time()

    def checkpoint(self, name=None):
        name = name if name is not None else Context().action_id
        now = time.time()
        self.timings.append(
            (name, now, now - self.relative_base, Context().action_duration)
        )

    @property
    def profile(self):
        profile = {}

        for _, timing in enumerate(self.timings):
            name, stamp, relative, duration = timing
            profile[name] = (stamp, f"{relative:.3f}s", f"{duration:.3f}s")

        return profile


class BaseContext:
    """
    A simple interpreter local context that can be used as a shared memory
    space without relying on "global" or careful variable passing.
    """
    def __init__(
        self,
        app_name: Optional[str]=None,
        config_path: Optional[str]=None,
        reinitialize: bool=False
    ):
        # skip duplicate initialization calls
        if not self._init and not reinitialize:
            return

        # constant vars
        self.app_name = app_name if app_name is not None else "coppyr"
        self.pid = os.getpid()
        self.config_path = config_path

        # dynamic vars
        self.ids = cycle(100000, 1000000)
        self.action_id = f"{self.pid}_{next(self.ids)}"
        self.action_stamp = time.time()
        self.logging_keys = [
            "action_id",
            "action_duration",
            "action_duration_str"
        ]

        # utility namespace
        self.namespace = Namespace()

        super().__init__()

    def setup(self):
        """
        Helper method that ensure some order dependent things are lazy loaded
        in the correct order.
        """
        self.config
        self.log

    @lazyproperty
    def app_name_env(self):
        return self.app_name.replace("-", "_")\
                            .replace(".", "__")\
                            .upper()

    @lazyproperty
    def ident(self):
        import binascii
        random_hex = binascii.b2a_hex(os.urandom(4)).decode()
        return random_hex

    @lazyproperty
    def uuid(self):
        import uuid
        return uuid.uuid4()

    @lazyproperty
    def host_id(self):
        import uuid
        import platform

        return uuid.uuid5(
            uuid.NAMESPACE_DNS, platform.node() + str(uuid.getnode())
        ).hex

    @lazyproperty
    def cwd(self):
        return os.getcwd()

    def inc_action_id(self):
        self.action_id = f"{self.ident}_{next(self.ids)}"  # increment action_id
        self.action_stamp = time.time()  # reset start time for new action

    @property
    def action_duration(self):
        return time.time() - self.action_stamp

    @property
    def action_duration_str(self):
        return f"{self.action_duration:.3f}"

    @lazyproperty
    def hostname(self):
        import socket
        return socket.gethostname()

    @lazyproperty
    def fqdn(self):
        import socket
        return socket.getfqdn()

    @lazyproperty
    def config(self):
        if self.config_path:
            from coppyr.config import YAMLConfig
            return YAMLConfig(self.config_path)
        else:
            return DotDict()

    @lazyproperty
    def log(self):
        from coppyr import logger

        if self.config.get("LOG_SETUP", True):
            logger.setup()

        if self.config.LOG_DISABLED:
            return logger.get("devnull")
        elif self.config.LOG_STDOUT:
            return logger.get("stdout", level=self.config.LOG_LEVEL)
        else:
            return logger.get(level=self.config.LOG_LEVEL)

    def get_file_handlers(self):
        # This helper is used by python-daemon to preserve file handlers when
        # starting a daemonized process.
        fds = [
            self.log.handlers[0].stream  # log file handler
        ]

        return fds

    @lazyproperty
    def timer(self):
        return TimeKeeper()


class Context(BaseContext, Singleton):
    pass



# -*- coding: utf-8 -*-
from contextlib import contextmanager


@contextmanager
def expect_exception(exc=Exception):
    """
    Provide context wrapper that wraps code and only passes if the specified
    exception is raised.
    """
    try:
        yield
        assert False  # exception was not raised
    except exc:
        # exception was raised, and we catch it here
        pass



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

from coppyr import (__title__, __summary__, __url__, __version__, __author__,
                    __email__, __license__)
from coppyr import package as pkg


setup(
    name=__title__,
    version=__version__,
    description=__summary__,
    long_description=pkg.get_readme("README.rst"),
    long_description_content_type="text/x-rst; charset=UTF-8",
    author=__author__,
    author_email=__email__,
    url=__url__,
    packages=find_packages(exclude=["tests", "tests.*"]),
    license=__license__,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    cmdclass={"upload": pkg.UploadCommand},
    install_requires=[],
    extras_require=pkg.parse_extras(
        config="/opt/coppyr/requirements-config.txt",
        daemon="/opt/coppyr/requirements-daemon.txt",
        dev="/opt/coppyr/requirements-dev.txt",
        pkg="/opt/coppyr/requirements-pkg.txt"
    ),
    include_package_data=True  # Read MANIFEST.in
)



import asyncio

from coppyr.collections import DotDict


class Loop:
    def __init__(self, new_event_loop=False):
        if new_event_loop:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        self.event_loop = asyncio.get_event_loop()
        self.tasks = set()
        self.results = []
        self.queues = DotDict()

    def task_complete(self, coroutine):
        self.tasks.discard(coroutine)
        self.results.append(coroutine.result())

    def create_task(self, coroutine, *args, **kwargs):
        # create the task
        task = self.event_loop.create_task(coroutine, *args, **kwargs)

        # add the tasks to tracking
        self.tasks.add(task)

        # on done, have the task remove itself
        task.add_done_callback(self.task_complete)

    def run_until_complete(self):
        self.event_loop.run_until_complete(asyncio.wait(self.tasks))

    def run_forever(self):
        self.event_loop.run_forever()

    def stop(self):
        self.event_loop.stop()

    def close(self):
        self.event_loop.close()

    def create_queue(self, name, maxsize=0):
        self.queues[name] = asyncio.Queue(maxsize=maxsize)
        return self.queues[name]

    def current_task(self):
        return asyncio.current_task()

    def is_running(self):
        return self.event_loop.is_running()

    def is_closed(self):
        return self.event_loop.is_closed()

